/*
 * Copyright (c) 2016-2017,2021-2022 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Jul  6 2016
 */
#include "acdynamixel.h"

#include <sys/time.h>
#include <err.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "dynamixel_c_types.h"
#include "codels.h"


const struct dx_modelid dx_models[] = {
  { .id = 0x1e, .name = "MX-28",
    .resolution = {
      .position = M_PI/2048,
      .velocity = 0.229/60,
      .acceleration = 214.577/60/60,
      .torque = 3.36/1000
    }
  },
  { .id = 0x137, .name = "MX-64",
    .resolution = {
      .position = M_PI/2048,
      .velocity = 0.229/60,
      .acceleration = 214.577/60/60,
      .torque = 3.36/1000
    }
  },
  { .id = 0x3fc, .name = "XM430-W350",
    .resolution = {
      .position = M_PI/2048,
      .velocity = 0.229/60,
      .acceleration = 214.577/60/60,
      .torque = 2.69/1000
    }
  },
  { .id = 0, .name = NULL }
};


/* --- Task comm -------------------------------------------------------- */


/** Codel dx_comm_start of task comm.
 *
 * Triggered by dynamixel_start.
 * Yields to dynamixel_configure.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_start(dynamixel_ids *ids, const dynamixel_motors *motors,
              const genom_context self)
{
  or_joint_state *mdata;
  int i;

  ids->chan = malloc(sizeof(*ids->chan));
  if (!ids->chan) return dx_e_sys_error(NULL, self);

  ids->chan->fd = -1;
  ids->chan->configured = false;
  memset(ids->chan->present, 0, sizeof(ids->chan->present));
  memset(ids->chan->expect, 0, sizeof(ids->chan->expect));
  ids->chan->r = ids->chan->w = 0;

  ids->update.update = false;
  for(i = 0; i < or_joint_max_joints; i++) {
    ids->update.limits[i] = false;
    ids->update.offset[i] = false;
    ids->update.profile[i] = false;
    ids->update.position_pidff[i] = false;
    ids->update.velocity_pi[i] = false;
  }

  for(i = 0; i < or_joint_max_joints; i++) {
    ids->resolution[i].position = nan("");
    ids->resolution[i].velocity = nan("");
    ids->resolution[i].acceleration = nan("");
    ids->resolution[i].torque = nan("");
  }

  ids->limits._length = 0;
  for(i = 0; i < ids->limits._maximum; i++)
    ids->limits._buffer[i] = (struct dynamixel_limits_s){
      .temperature_max = nan(""),
      .voltage_min = nan(""), .voltage_max = nan(""), .pwm_max = nan(""),
      .current_max = nan(""), .acceleration_max = nan(""),
      .velocity_max = nan("")
    };

  ids->offset._length = 0;
  for(i = 0; i < ids->offset._maximum; i++)
    ids->offset._buffer[i] = nan("");

  ids->profile._length = 0;
  for(i = 0; i < ids->profile._maximum; i++)
    ids->profile._buffer[i] = (struct dynamixel_profile_s){
      .velocity = nan(""), .acceleration = nan("")
    };
  ids->position_pidff._length = 0;
  for(i = 0; i < ids->position_pidff._maximum; i++)
    ids->position_pidff._buffer[i] = (struct dynamixel_pidff_s){
      .kP = nan(""), .kI = nan(""), .kD = nan(""),
      .kFF1 = nan(""), .kFF2 = nan("")
    };
  ids->velocity_pi._length = 0;
  for(i = 0; i < ids->velocity_pi._maximum; i++)
    ids->velocity_pi._buffer[i] = (struct dynamixel_pi_s){
      .kP = nan(""), .kI = nan("")
    };

  ids->input.mode = dynamixel_cmd_position;
  ids->input.position._length = 0;
  for(i = 0; i < ids->input.position._maximum; i++)
    ids->input.position._buffer[i] = nan("");
  ids->input.velocity._length = 0;
  for(i = 0; i < ids->input.velocity._maximum; i++)
    ids->input.position._buffer[i] = nan("");
  ids->input.effort._length = 0;
  for(i = 0; i < ids->input.effort._maximum; i++)
    ids->input.effort._buffer[i] = nan("");

  mdata = motors->data(self);
  if (mdata) {
    mdata->ts.sec = mdata->ts.nsec = 0;
    mdata->name._length = 0;
    mdata->flag._length = mdata->flag._maximum;
    mdata->position._length = 0;
    mdata->velocity._length = 0;
    mdata->effort._length = 0;
    for(i = 0; i < mdata->flag._maximum; i++)
      mdata->flag._buffer[i].active = false;
  }

  return dynamixel_configure;
}


/** Codel dx_comm_configure of task comm.
 *
 * Triggered by dynamixel_configure.
 * Yields to dynamixel_pause_configure, dynamixel_poll.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_configure(const dynamixel_channel_s *chan,
                  dynamixel_ids_update_s *update,
                  dynamixel_cmd_mode *mode, const genom_context self)
{
  if (chan->fd < 0) return dynamixel_pause_configure;

  /* request motor parameters update */
  update->update = true;

  /* disable status for writes */
  dx_send_msg(chan, 0xfe, DX_WRITE, 1, DX_STATUS_RETURN_LEVEL, 1);

  /* release motor */
  dx_send_msg(chan, 0xfe, DX_WRITE, 1, DX_TORQUE_ENABLE, 0);
  dx_send_msg(chan, 0xfe, DX_WRITE, 1, DX_LED, 0);

  /* disable delay */
  dx_send_msg(chan, 0xfe, DX_WRITE, 1, DX_RETURN_DELAY_TIME, 0);

  /* disable secondary id */
  dx_send_msg(chan, 0xfe, DX_WRITE, 1, DX_SECONDARY_ID, 0xff);

  /* unlimit hardware */
  dx_send_msg(chan, 0xfe, DX_WRITE, 2, DX_PWM_LIMIT, 0x75, 0x3);

  dx_send_msg(chan, 0xfe, DX_WRITE, 4, DX_ACCELERATION_LIMIT,
              0xff, 0x7f, 0, 0);
  dx_send_msg(chan, 0xfe, DX_WRITE, 4, DX_VELOCITY_LIMIT, 0xff, 0x3, 0, 0);

  /* set position control mode */
  dx_send_msg(chan, 0xfe, DX_WRITE, 4, DX_GOAL_VELOCITY, 0, 0, 0, 0);
  dx_send_msg(chan, 0xfe, DX_WRITE, 2, DX_GOAL_CURRENT, 0, 0);
  *mode = dynamixel_cmd_position;
  dx_send_operating_mode(chan, 0xfe, DX_EXT_POSITION_MODE);

  /* get model id */
  dx_send_msg(chan, or_joint_max_joints, DX_SYNC_READ, DX_ID, 0);

  return dynamixel_poll;
}


/** Codel dx_comm_update of task comm.
 *
 * Triggered by dynamixel_update.
 * Yields to dynamixel_poll.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_update(const dynamixel_channel_s *chan,
               dynamixel_ids_update_s *update,
               const dynamixel_resolution_s resolution[8],
               const sequence8_dynamixel_limits_s *limits,
               const sequence8_double *offset,
               const sequence8_dynamixel_profile_s *profile,
               const sequence8_dynamixel_pidff_s *position_pidff,
               const sequence8_dynamixel_pi_s *velocity_pi,
               const genom_context self)
{
  int i;

  /* update limits */
  for(i = 0; i < limits->_length; i++)
    if (update->limits[i]) {
      dx_send_limits(
        chan, i, &limits->_buffer[i], &resolution[i]);
      update->limits[i] = false;
    }

  /* update homing offset */
  for(i = 0; i < offset->_length; i++)
    if (update->offset[i]) {
      dx_send_homing_offset(
        chan, i, offset->_buffer[i], &resolution[i]);
      update->offset[i] = false;
    }

  /* update controller parameters */
  for(i = 0; i < profile->_length; i++) {
    if (update->profile[i]) {
      dx_send_profile(chan, i, &profile->_buffer[i], &resolution[i]);
      update->profile[i] = false;
    }
  }
  for(i = 0; i < position_pidff->_length; i++) {
    if (update->position_pidff[i]) {
      dx_send_position_pidff(chan, i, &position_pidff->_buffer[i]);
      update->position_pidff[i] = false;
    }
  }
  for(i = 0; i < velocity_pi->_length; i++) {
    if (update->velocity_pi[i]) {
      dx_send_velocity_pi(chan, i, &velocity_pi->_buffer[i]);
      update->velocity_pi[i] = false;
    }
  }

  update->update = false;

  /* re-read configuration (homing offset, controller parameters) */
  dx_send_msg(chan, or_joint_max_joints, DX_SYNC_READ,
              DX_PROFILE_VELOCITY+4, 0);

  return dynamixel_poll;
}


/** Codel dx_comm_send of task comm.
 *
 * Triggered by dynamixel_send.
 * Yields to dynamixel_pause_configure, dynamixel_poll.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_send(const dynamixel_channel_s *chan,
             const dynamixel_resolution_s resolution[8],
             dynamixel_ids_input_s *input,
             const sequence8_dynamixel_profile_s *profile,
             const genom_context self)
{
  int i;

  if (!chan->configured) return dynamixel_pause_configure;

  /* send position target */
  if (input->position._length > 0) {
    uint8_t values[4 * input->position._length], *v;
    int32_t value;

    if (input->effort._length > 0) {
      if (input->mode != dynamixel_cmd_position_effort) {
        dx_send_operating_mode(chan, 0xfe, DX_TORQUE_POSITION_MODE);

        /* when switching mode, profile is reset */
        for(i = 0; i < profile->_length; i++)
          dx_send_profile(chan, i, &profile->_buffer[i], &resolution[i]);

        input->mode = dynamixel_cmd_position_effort;
      }
    } else if (input->mode != dynamixel_cmd_position) {
      dx_send_operating_mode(chan, 0xfe, DX_EXT_POSITION_MODE);

      /* when switching mode, profile is reset */
      for(i = 0; i < profile->_length; i++)
        dx_send_profile(chan, i, &profile->_buffer[i], &resolution[i]);

      input->mode = dynamixel_cmd_position;
    }

    /* send position target */
    v = values;
    for(i = 0; i < input->position._length; i++) {
      value = input->position._buffer[i] / resolution[i].position;
      *v++ = value & 0xff;
      *v++ = (value >> 8) & 0xff;
      *v++ = (value >> 16) & 0xff;
      *v++ = (value >> 24) & 0xff;
    }

    dx_send_msg(chan, input->position._length, DX_SYNC_WRITE,
                4, DX_GOAL_POSITION, values);
  }


  /* send velocity target */
  if (input->velocity._length > 0) {
    uint8_t values[4 * input->velocity._length], *v;
    int32_t value;

    if (input->mode != dynamixel_cmd_velocity) {
      dx_send_operating_mode(chan, 0xfe, DX_VELOCITY_MODE);

      /* when switching mode, profile is reset */
      for(i = 0; i < profile->_length; i++)
        dx_send_profile(chan, i, &profile->_buffer[i], &resolution[i]);

      input->mode = dynamixel_cmd_velocity;
    }

    v = values;
    for(i = 0; i < input->velocity._length; i++) {
      value = input->velocity._buffer[i] / resolution[i].velocity;

      *v++ = value & 0xff;
      *v++ = (value >> 8) & 0xff;
      *v++ = (value >> 16) & 0xff;
      *v++ = (value >> 24) & 0xff;
    }

    dx_send_msg(chan, input->velocity._length, DX_SYNC_WRITE,
                  4, DX_GOAL_VELOCITY, values);
  }


  /* send effort target */
  if (input->effort._length > 0) {
    uint8_t values[2 * input->effort._length], *v;
    int16_t value;

    if (input->position._length == 0 &&
        input->mode != dynamixel_cmd_effort) {
      dx_send_operating_mode(chan, 0xfe, DX_TORQUE_MODE);

      /* when switching mode, profile is reset */
      for(i = 0; i < profile->_length; i++)
        dx_send_profile(chan, i, &profile->_buffer[i], &resolution[i]);

      input->mode = dynamixel_cmd_effort;
    }

    v = values;
    for(i = 0; i < input->effort._length; i++) {
      value = input->effort._buffer[i] / resolution[i].torque;

      *v++ = value & 0xff;
      *v++ = (value >> 8) & 0xff;
    }

    dx_send_msg(chan, input->effort._length, DX_SYNC_WRITE,
                2, DX_GOAL_CURRENT, values);
  }

  /* request data from all motors */
  dx_send_msg(chan, or_joint_max_joints, DX_SYNC_READ,
              10, DX_PRESENT_CURRENT);

  /* reset input - no need to send the same twice */
  input->position._length = 0;
  input->velocity._length = 0;
  input->effort._length = 0;

  return dynamixel_poll;
}


/** Codel dx_comm_poll of task comm.
 *
 * Triggered by dynamixel_poll.
 * Yields to dynamixel_pause_configure, dynamixel_recv.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_poll(const dynamixel_channel_s *chan,
             const genom_context self)
{
  int s;

  s = dx_wait_msg(chan, 250);
  switch(s) {
    case -2:
      warnx("disconnected from %s", chan->path);
      close(chan->fd);
      ((dynamixel_channel_s *)chan)->fd = -1;
      ((dynamixel_channel_s *)chan)->configured = false;
      memset(((dynamixel_channel_s *)chan)->present, 0, sizeof(chan->present));
      memset(((dynamixel_channel_s *)chan)->expect, 0, sizeof(chan->expect));
      return dynamixel_pause_configure;

    case -1: case 0:
      if (chan->configured) {
        warnx("%s became inactive", chan->path);
        ((dynamixel_channel_s *)chan)->configured = false;
      }
      return dynamixel_pause_configure;
  }
  return dynamixel_recv;
}


/** Codel dx_comm_recv of task comm.
 *
 * Triggered by dynamixel_recv.
 * Yields to dynamixel_pause_update, dynamixel_pause_send.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_recv(dynamixel_channel_s **chan,
             const dynamixel_ids_update_s *update,
             dynamixel_resolution_s resolution[8],
             sequence8_dynamixel_limits_s *limits,
             sequence8_dynamixel_profile_s *profile,
             sequence8_dynamixel_pidff_s *position_pidff,
             sequence8_dynamixel_pi_s *velocity_pi,
             sequence8_double *offset, const dynamixel_motors *motors,
             const genom_context self)
{
  or_joint_state *mdata = motors->data(self);
  struct timeval tv;
  int32_t v32;
  int16_t v16;
  int id;

  gettimeofday(&tv, NULL);
  for(id = 0; id < mdata->flag._length; id++)
    mdata->flag._buffer[id-1].active = false;
  mdata->position._length = 0;
  mdata->velocity._length = 0;
  mdata->effort._length = 0;

  while(dx_recv_msg(*chan, false) == 1) {
    id = (*chan)->id;
    if (id < 1 || id > or_joint_max_joints) {
      warnx("dropping motor %d: id greater than %d", id, or_joint_max_joints);
      continue;
    }
    id--;

    switch ((*chan)->p) {
      case 10: /* motor state request from dx_comm_send */
        mdata->ts.sec = tv.tv_sec;
        mdata->ts.nsec = 1000 * tv.tv_usec;

        mdata->flag._buffer[id].active = true;

        if (mdata->position._length <= id) mdata->position._length = id + 1;
        if (mdata->velocity._length <= id) mdata->velocity._length = id + 1;
        if (mdata->effort._length <= id) mdata->effort._length = id + 1;

        v32 = (*chan)->param[6];
        v32 |= ((uint32_t)(*chan)->param[7] << 8);
        v32 |= ((uint32_t)(*chan)->param[8] << 16);
        v32 |= ((int32_t)(*chan)->param[9] << 24);
        mdata->position._buffer[id] = v32 * resolution[id].position;

        v32 = (*chan)->param[2];
        v32 |= ((uint32_t)(*chan)->param[3] << 8);
        v32 |= ((uint32_t)(*chan)->param[4] << 16);
        v32 |= ((int32_t)(*chan)->param[5] << 24);
        mdata->velocity._buffer[id] = v32 * resolution[id].velocity;

        v16 = (*chan)->param[0];
        v16 |= ((int16_t)(*chan)->param[1] << 8);
        mdata->effort._buffer[id] = v16 * resolution[id].torque;
        break;

      case DX_ID: { /* motor id request from dx_comm_configure */
        int model;
        int i;

        if (mdata->name._length <= id) mdata->name._length = id + 1;
        snprintf(mdata->name._buffer[id],
                 sizeof(mdata->name._buffer[id]), "unknown (version %d)",
                 (*chan)->param[2]);
        resolution[id].position = nan("");
        resolution[id].velocity = nan("");
        resolution[id].acceleration = nan("");
        resolution[id].torque = nan("");

        model =
          (*chan)->param[DX_MODEL_NUMBER] +
          ((*chan)->param[DX_MODEL_NUMBER+1] << 8);
        for(i = 0; dx_models[i].id; i++) {
          if (dx_models[i].id == model) {
            snprintf(mdata->name._buffer[id],
                     sizeof(mdata->name._buffer[id]), "%s (version %d)",
                     dx_models[i].name, (*chan)->param[DX_FIRMWARE_VERSION]);
            resolution[id] = dx_models[i].resolution;
            break;
          }
        }

        warnx("detected motor %d: %s", id + 1, mdata->name._buffer[id]);
        break;
      }

      case DX_PROFILE_VELOCITY+4: { /* parameters request from dx_comm_update */
        /* homing offset */
        if (!update->offset[id]) {
          if (offset->_length <= id) offset->_length = id + 1;
          dx_recv_homing_offset(
            (*chan)->param, &offset->_buffer[id], &resolution[id]);
        }

        /* limits */
        if (!update->limits[id]) {
          if (limits->_length <= id) limits->_length = id + 1;
          dx_recv_limits(
            (*chan)->param, &limits->_buffer[id], &resolution[id]);
        }

        /* controller parameters */
        if (!update->profile[id]) {
          if (profile->_length <= id) profile->_length = id + 1;
          dx_recv_profile((*chan)->param, &profile->_buffer[id],
                          &resolution[id]);
        }
        if (!update->position_pidff[id]) {
          if (position_pidff->_length <= id) position_pidff->_length = id + 1;
          dx_recv_position_pidff((*chan)->param,
                                 &position_pidff->_buffer[id]);
        }
        if (!update->velocity_pi[id]) {
          if (velocity_pi->_length <= id) velocity_pi->_length = id + 1;
          dx_recv_velocity_pi((*chan)->param, &velocity_pi->_buffer[id]);
        }

        (*chan)->configured = true;
        break;
      }

      default:
        warnx("spurious message length %d", (*chan)->p);
    }
  }
  motors->write(self);

  return update->update ? dynamixel_pause_update : dynamixel_pause_send;
}


/** Codel dx_comm_stop of task comm.
 *
 * Triggered by dynamixel_stop.
 * Yields to dynamixel_ether.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_comm_stop(dynamixel_channel_s **chan, const genom_context self)
{
  return dx_disconnect_start(chan, self);
}


/* --- Activity connect ------------------------------------------------- */

/** Codel dx_connect_start of activity connect.
 *
 * Triggered by dynamixel_start.
 * Yields to dynamixel_wait.
 * Throws dynamixel_e_sys, dynamixel_e_baddev.
 */
genom_event
dx_connect_start(const char serial[64], uint32_t baud,
                 dynamixel_channel_s **chan, const genom_context self)
{
  int s, c;

  if ((*chan)->fd >= 0) {
    (*chan)->configured = false;
    close((*chan)->fd);
    warnx("disconnected from %s", (*chan)->path);
  }

  /* open tty */
  (*chan)->fd = dx_open_tty(serial, baud);
  if ((*chan)->fd < 0) return dx_e_sys_error(serial, self);
  snprintf((*chan)->path, sizeof((*chan)->path), "%s", serial);
  (*chan)->state = 0;
  (*chan)->header = 0;

  while (dx_recv_msg(*chan, false) == 1); /* flush buffer */

  /* check endpoint */
  c = 0;
  do {
    if (dx_send_msg(*chan, 0xfe, DX_PING)) {
      close((*chan)->fd);
      (*chan)->fd = -1;
      (*chan)->configured = false;
      return dx_e_sys_error(serial, self);
    }

    s = dx_wait_msg(*chan, 250/*ms*/);
    if (s < 0) {
      close((*chan)->fd);
      (*chan)->fd = -1;
      (*chan)->configured = false;
      return dx_e_sys_error(NULL, self);
    } else if (s == 0 && c > 1) {
      /* broadcast a protocol 1.0 message switching to protocol 2,
       * just in case (write reg 0xd to 0x2 on id 0xfe) */
      if (write((*chan)->fd, (uint8_t []){
            0xff, 0xff, 0xfe, 0x04, 0x03, 0x0d, 0x02, 0xeb
            }, 8))
        /* ignore errors */;
    } else if (s > 0)
      s = dx_recv_msg(*chan, true);

  } while(s != 1 && c++ < 3);
  while (dx_recv_msg(*chan, true) == 1); /* flush buffer */

  if (c > 3) {
    close((*chan)->fd);
    (*chan)->fd = -1;
    (*chan)->configured = false;
    errno = ETIMEDOUT;
    return dx_e_sys_error(NULL, self);
  }

  return dynamixel_wait;
}

/** Codel dx_connect_wait of activity connect.
 *
 * Triggered by dynamixel_wait.
 * Yields to dynamixel_pause_wait, dynamixel_ether.
 * Throws dynamixel_e_sys, dynamixel_e_baddev.
 */
genom_event
dx_connect_wait(const dynamixel_channel_s *chan,
                const genom_context self)
{
  if (!chan->configured) return dynamixel_pause_wait;

  warnx("connected to %s", chan->path);
  return dynamixel_ether;
}


/* --- Activity disconnect ---------------------------------------------- */

/** Codel dx_disconnect_start of activity disconnect.
 *
 * Triggered by dynamixel_start.
 * Yields to dynamixel_ether.
 * Throws dynamixel_e_sys.
 */
genom_event
dx_disconnect_start(dynamixel_channel_s **chan,
                    const genom_context self)
{
  /* set 0 velocity */
  uint8_t z[4 * or_joint_max_joints];

  memset(z, 0, sizeof(z));
  dx_send_operating_mode(*chan, 0xfe, DX_VELOCITY_MODE);
  dx_send_msg(*chan, or_joint_max_joints, DX_SYNC_WRITE,
              4, DX_GOAL_VELOCITY, z);


  /* release motor */
  dx_send_msg(*chan, 0xfe, DX_WRITE, 1, DX_TORQUE_ENABLE, 0);
  dx_send_msg(*chan, 0xfe, DX_WRITE, 1, DX_LED, 0);

  /* close serial */
  if ((*chan)->fd >= 0) {
    close((*chan)->fd);
    (*chan)->fd = -1;
    (*chan)->configured = false;
    warnx("disconnected from %s", (*chan)->path);
  }

  return dynamixel_ether;
}


/* --- Activity servo --------------------------------------------------- */

/** Codel dx_servo_start of activity servo.
 *
 * Triggered by dynamixel_start.
 * Yields to dynamixel_pause_start, dynamixel_stop.
 * Throws dynamixel_e_sys, dynamixel_e_input.
 */
genom_event
dx_servo_start(dynamixel_ids_input_s *input,
               const dynamixel_joint_input *joint_input,
               const genom_context self)
{
  or_joint_input *input_data;
  struct timeval tv;
  uint32_t i;

  /* update input */
  if (joint_input->read(self)) return dynamixel_e_input(self);

  input_data = joint_input->data(self);
  if (!input_data) return dynamixel_e_input(self);

  /* watchdog on input */
  gettimeofday(&tv, NULL);
  if (tv.tv_sec + 1e-6*tv.tv_usec >
      0.5 + input_data->ts.sec + 1e-9*input_data->ts.nsec) {
    dx_stop(input, self);
    return dynamixel_e_input(self);
  }

  /* apply data */
#define dx_copy_data(x)                                                 \
  do {                                                                  \
    if (input_data->x._present) {                                       \
      for(i = 0; i < input_data->x._value._length; i++)                 \
        input->x._buffer[i] = input_data->x._value._buffer[i];          \
      input->x._length = input_data->x._value._length;                  \
    } else                                                              \
      input->x._length = 0;                                             \
  } while(0)

  dx_copy_data(position);
  dx_copy_data(velocity);
  dx_copy_data(effort);
#undef dx_copy_data

  if (input_data->position._present)
    if (input_data->effort._present)
      dx_set_position_effort_validate(
        &input->position, &input->velocity, &input->effort, self);
    else
      dx_set_position_validate(
        &input->position, &input->velocity, &input->effort, self);
  else if (input_data->velocity._present)
    dx_set_velocity_validate(
      &input->position, &input->velocity, &input->effort, self);
  else if (input_data->effort._present)
    dx_set_effort_validate(
      &input->position, &input->velocity, &input->effort, self);

  return dynamixel_pause_start;
}

/** Codel dx_servo_stop of activity servo.
 *
 * Triggered by dynamixel_stop.
 * Yields to dynamixel_ether.
 * Throws dynamixel_e_sys, dynamixel_e_input.
 */
genom_event
dx_servo_stop(dynamixel_ids_input_s *input, const genom_context self)
{
  dx_stop(input, self);
  return dynamixel_ether;
}


/* --- Activity configure ----------------------------------------------- */

/** Codel dx_configure_start of activity configure.
 *
 * Triggered by dynamixel_start.
 * Yields to dynamixel_scan.
 * Throws dynamixel_e_sys, dynamixel_e_badrate.
 */
genom_event
dx_configure_start(dynamixel_channel_s **lchan,
                   dynamixel_channel_s **chan,
                   const genom_context self)
{
  genom_event e;

  *lchan = malloc(sizeof(**lchan));
  if (!*lchan) return dx_e_sys_error(NULL, self);
  **lchan = (dynamixel_channel_s){ .configured = false, .fd = -1 };

  e = dx_disconnect_start(chan, self);
  if (e != dynamixel_ether) return e;

  return dynamixel_scan;
}

/** Codel dx_configure_scan of activity configure.
 *
 * Triggered by dynamixel_scan.
 * Yields to dynamixel_ether.
 * Throws dynamixel_e_sys, dynamixel_e_badrate.
 */
genom_event
dx_configure_scan(const char serial[64], dynamixel_channel_s **lchan,
                  uint16_t id, uint32_t baud,
                  const genom_context self)
{
  static const uint32_t bauds[] = {
    9600, 57600, 115200, 1000000, 2000000, 3000000
  };
  genom_event e;
  const char *mname;
  int model, i, j, b;

  for (b = 0; b < sizeof(bauds)/sizeof(bauds[0]); b++)
    if (bauds[b] == baud) { baud = b; break; }
  if (b >= sizeof(bauds)/sizeof(bauds[0])) return dynamixel_e_badrate(self);

  /* scan known baud rates */
  for (b = 0; b < sizeof(bauds)/sizeof(bauds[0]); b++) {
    warnx("scanning %s at %d bps", serial, bauds[b]);
    e = dx_connect_start(serial, bauds[b], lchan, self);
    if (e == dynamixel_wait) break;
  }
  if (e != dynamixel_wait) {
    free(*lchan);
    return e;
  }

  /* get motor info */
  dx_send_msg(*lchan, or_joint_max_joints, DX_SYNC_READ,
              14, DX_MODEL_NUMBER);
  while(dx_recv_msg(*lchan, true) == 1) {
    if ((*lchan)->p != 14) continue;
    i = (*lchan)->id;
    if (i >= or_joint_max_joints) continue;

    mname = "unknown model";
    model = (*lchan)->param[0] + ((*lchan)->param[1] << 8);
    for(j = 0; dx_models[j].id; j++) {
      if (dx_models[j].id == model) {
        mname = dx_models[j].name;
        break;
      }
    }

    /* configure */
    warnx("detected motor %d: %s, version %d @ %d bps",
          i, mname, (*lchan)->param[6], bauds[(*lchan)->param[8]]);

    /* disable status for writes */
    dx_send_msg(*lchan, i, DX_WRITE, 1, DX_STATUS_RETURN_LEVEL, 1);

    /* release motor */
    dx_send_msg(*lchan, i, DX_WRITE, 1, DX_TORQUE_ENABLE, 0);
    dx_send_msg(*lchan, i, DX_WRITE, 1, DX_LED, 0);

    /* disable delay */
    dx_send_msg(*lchan, i, DX_WRITE, 1, DX_RETURN_DELAY_TIME, 0);

    /* set id, baudrate */
    dx_send_msg(*lchan, i, DX_WRITE, 2, DX_ID, id, baud);

    /* reboot */
    dx_send_msg(*lchan, id, DX_REBOOT);

    warnx("reconfigured motor %d: id %d @ %d bps", i, id, bauds[baud]);
  }

  close((*lchan)->fd);
  free(*lchan);
  return dynamixel_ether;
}
